REM ----------------------------------------
REM Video emulation core of a GameBoy device
REM This file is part of the zBoy project
REM Copyright (C) Mateusz Viste 2010
REM ----------------------------------------
REM
REM FF40 (LCDC register):
REM Bit7 0=Display Off  1=Display On
REM Bit6 0=TileMap at 9800-9BFF   1=TileMap at 9C00-9FFF
REM Bit5 0=WindowDisplay disabled  1=WindowDisplay enabled
REM Bit4 0=BG&Window tile data at 8800-97FF   1=BG&Window tile data at 8000-8FFF
REM Bit3 0=BgTile Map display at 9800-9BFF   1=BgTitle Map display at 9C00-9FFF
REM Bit2 0=SpriteSize=8x8    1=SpriteSize=8x16
REM Bit1 0=Do not display Sprites   1=Display sprites
REM Bit0 0=Do not display background   1=Display background
REM
REM Generating an interrupt means "enabling the corresponding bit of the IF register (FF0F)"
REM

DIM SHARED AS UBYTE HideBackgroundDisplay = 0
DIM SHARED AS UBYTE HideSpritesDisplay = 0

DIM SHARED AS DOUBLE LastScreenshotTime
DIM SHARED AS UBYTE AutoScreenshot = 0

DIM SHARED AS UINTEGER VideoClkCounterMode, VideoClkCounterVBlank, CountFPS
DIM SHARED ScreenBuffer(0 TO 159, 0 TO 143) AS UBYTE
DIM SHARED ScreenOldBuffer(0 TO 159, 0 TO 143) AS UBYTE  ' used as a buffer to detect changes on screen
DIM SHARED AS UBYTE LcdModifiedFirstY, LcdModifiedLastY, LcdModifiedFirstX, LcdModifiedLastX
DIM SHARED LcdModifiedScanline(0 TO 143) AS UBYTE
DIM SHARED AS UBYTE SkipFrames = 1, SkipFramesCounter = 1
DIM SHARED AS UBYTE LastLY, CurLY, LastLYdraw  ' used by VideoSysUpdate
DIM SHARED ScreenPalette32(0 TO 15) AS UINTEGER  ' Palette of shades to use onscreen
ScreenPalette32(&h0) = RGB(&hFF,&hFF,&hFF)    ' Brightest shade
ScreenPalette32(&h01) = RGB(&hAA,&hAA,&hAA)
ScreenPalette32(&h02) = RGB(&h55,&h55,&h55)
ScreenPalette32(&h03) = RGB(&h00,&h00,&h00)    ' Darkest shade
ScreenPalette32(&h0E) = RGB(&h00,&h00,&h00)    ' This one is used as screen's borders in fullscreen mode.
ScreenPalette32(&h0F) = RGB(&h30,&h30,&h30)    ' This one is used as background when FastScale is used.
DIM SHARED ScreenPalette16(0 TO 15) AS USHORT ' Palette of shades to use onscreen
ScreenPalette16(&h00) = &b1111111111111111     '  R=11111 G=111111 B=11111
ScreenPalette16(&h01) = &b1010110101010101     '  R=10101 G=101010 B=10101
ScreenPalette16(&h02) = &b0101001010101010     '  R=01010 G=010101 B=01010
ScreenPalette16(&h03) = &b0000000000000000     '  R=00000 G=000000 B=00000
ScreenPalette16(&h0E) = &b0000000000000000     '  R=00000 G=000000 B=00000
ScreenPalette16(&h0F) = &b0011000110000110     '  R=00110 G=001100 B=00110

#INCLUDE ONCE "hqvideo.bi"


'FUNCTION GbPalette(BYREF ColIdx AS UBYTE, BYREF PalType AS UBYTE) AS UBYTE
'  REM  Returns the RGB value of a color index from one of the palettes
'  REM   ColIdx: 0..3
'  REM   PalType: 0=BGP  1=OBP0  2=OBP1
'  SELECT CASE ColIdx
'    CASE 0
'      RETURN (IoRegisters(&hFF47 + PalType) AND &b00000011)
'    CASE 2
'      RETURN ((IoRegisters(&hFF47 + PalType) AND &b00001100) SHR 2)
'    CASE 1
'      RETURN ((IoRegisters(&hFF47 + PalType) AND &b00110000) SHR 4)
'    CASE 3
'      RETURN ((IoRegisters(&hFF47 + PalType) AND &b11000000) SHR 6)
'  END SELECT
'END FUNCTION

#MACRO GbPalette0(ColIdx, ReturnValue)
  REM  Returns the RGB value of a color index from one of the palettes
  REM   ColIdx: 0..3
  REM   PalType: 0=BGP  1=OBP0  2=OBP1
  SELECT CASE ColIdx
    CASE 0
      ReturnValue = (IoRegisters(&hFF47) AND &b00000011)
    CASE 2
      ReturnValue = ((IoRegisters(&hFF47) AND &b00001100) SHR 2)
    CASE 1
      ReturnValue = ((IoRegisters(&hFF47) AND &b00110000) SHR 4)
    CASE 3
      ReturnValue = ((IoRegisters(&hFF47) AND &b11000000) SHR 6)
  END SELECT
#ENDMACRO
#MACRO GbPalette1(ColIdx, ReturnValue)
  REM  Returns the RGB value of a color index from one of the palettes
  REM   ColIdx: 0..3
  REM   PalType: 0=BGP  1=OBP0  2=OBP1
  SELECT CASE ColIdx
    CASE 0
      ReturnValue = (IoRegisters(&hFF48) AND &b00000011)
    CASE 2
      ReturnValue = ((IoRegisters(&hFF48) AND &b00001100) SHR 2)
    CASE 1
      ReturnValue = ((IoRegisters(&hFF48) AND &b00110000) SHR 4)
    CASE 3
      ReturnValue = ((IoRegisters(&hFF48) AND &b11000000) SHR 6)
  END SELECT
#ENDMACRO
#MACRO GbPalette2(ColIdx, ReturnValue)
  REM  Returns the RGB value of a color index from one of the palettes
  REM   ColIdx: 0..3
  REM   PalType: 0=BGP  1=OBP0  2=OBP1
  SELECT CASE ColIdx
    CASE 0
      ReturnValue = (IoRegisters(&hFF49) AND &b00000011)
    CASE 2
      ReturnValue = ((IoRegisters(&hFF49) AND &b00001100) SHR 2)
    CASE 1
      ReturnValue = ((IoRegisters(&hFF49) AND &b00110000) SHR 4)
    CASE 3
      ReturnValue = ((IoRegisters(&hFF49) AND &b11000000) SHR 6)
  END SELECT
#ENDMACRO



'SUB DrawLine(BYREF CurScanline AS UBYTE)
SUB DrawBackground(BYREF CurScanline AS UBYTE)
  STATIC AS INTEGER TilesDataAddr, BgTilesMapAddr, TileNum, TileToDisplay, TileTempAddress
  STATIC AS UBYTE x, y, z, t, u, UbyteBuff1, UbyteBuff2, PixelX
  STATIC TileBuffer(0 TO 63) AS UBYTE
  IF ((IoRegisters(&hFF40) AND &b00000001) = 0) OR (HideBackgroundDisplay = 1) THEN ' if "BackgroundEnabled" bit is not set, or bg has been forced OFF by user, then do not draw background (fill with black instead)
      FOR x = 0 TO 159
        ScreenBuffer(x, CurScanline) = 3 ' black
      NEXT x
    ELSE  ' If "BackgroundEnabled" bit is set then draw background
      REM Get starting address of tiles data
      IF (IoRegisters(&hFF40) AND &b00010000) = 0 THEN
          TilesDataAddr = &h8800
        ELSE
          TilesDataAddr = &h8000
      END IF
      IF (IoRegisters(&hFF40) AND &b00001000) = 0 THEN ' Get address of BGP tiles *map*
          BgTilesMapAddr = &h9800
        ELSE
          BgTilesMapAddr = &h9C00
      END IF
      y = (CurScanline + IoRegisters(&hFF42))
      z = (y AND &b0111)  ' Same than z = y MOD 8 (but MUCH faster)    --> this is the tile's row that has to be computed
      u = (z SHL 3)    ' SHL 3 = *8
      y = (y SHR 3) ' Same than y = y/8 (but SHR is faster)            --> this is the number of the tile to display

      'FOR y = 0 TO 31
        FOR x = 0 TO 31
          TileNum = (y SHL 5) + x    ' SHL 5 means "*32"
          IF TilesDataAddr = &h8000 THEN
              TileToDisplay = VideoRAM(BgTilesMapAddr + TileNum)
            ELSE
              TileToDisplay = UbyteToByte(VideoRAM(BgTilesMapAddr + TileNum)) + 128
          END IF
          t = (z SHL 1)   '      Get only the tile's y line related to current scanline
          TileTempAddress = TilesDataAddr + (TileToDisplay SHL 4) + t
          t = t SHL 2
          'FOR t = 0 TO 15 STEP 2
            UbyteBuff1 = VideoRAM(TileTempAddress)
            UbyteBuff2 = VideoRAM(TileTempAddress + 1)
            TileBuffer(t + 7) =  ((UbyteBuff2 AND &b00000001) OR ((UbyteBuff1 AND &b00000001) SHL 1))
            TileBuffer(t + 6) = (((UbyteBuff2 AND &b00000010) SHR 1) OR (((UbyteBuff1 AND &b00000010) SHR 1) SHL 1))
            TileBuffer(t + 5) = (((UbyteBuff2 AND &b00000100) SHR 2) OR (((UbyteBuff1 AND &b00000100) SHR 2) SHL 1))
            TileBuffer(t + 4) = (((UbyteBuff2 AND &b00001000) SHR 3) OR (((UbyteBuff1 AND &b00001000) SHR 3) SHL 1))
            TileBuffer(t + 3) = (((UbyteBuff2 AND &b00010000) SHR 4) OR (((UbyteBuff1 AND &b00010000) SHR 4) SHL 1))
            TileBuffer(t + 2) = (((UbyteBuff2 AND &b00100000) SHR 5) OR (((UbyteBuff1 AND &b00100000) SHR 5) SHL 1))
            TileBuffer(t + 1) = (((UbyteBuff2 AND &b01000000) SHR 6) OR (((UbyteBuff1 AND &b01000000) SHR 6) SHL 1))
            TileBuffer(t + 0) = (((UbyteBuff2 AND &b10000000) SHR 7) OR (((UbyteBuff1 AND &b10000000) SHR 7) SHL 1))
          'NEXT t
          PixelX = ((x SHL 3) - IoRegisters(&hFF43))  ' &hFF43 is the SCX register
          FOR t = 0 TO 7
            IF (PixelX < 160) THEN
              'ScreenBuffer(PixelX, CurScanline) = GbPalette(TileBuffer((z SHL 3) + t), 0)
              GbPalette0(TileBuffer(u + t), ScreenBuffer(PixelX, CurScanline))
            END IF
            PixelX += 1
          NEXT t
        NEXT x
      'NEXT y
  END IF
END SUB


SUB DrawWindow(BYREF CurScanline AS UBYTE)
  REM
  REM Some infos about the window...
  REM
  REM WindowPosX = 7    [FF4Bh]  (should be >= 0 and <= 166)
  REM WindowPosY = 0    [FF4Ah]  (should be >= 0 and <= 143)
  REM
  REM Window tiles can be located at 8000h-8FFFh or 8800h-97FFh (just like for the background)
  REM At 8000h-8FFFh tiles are numbered from 0 to 255, while at 8800h-97FFh tiles are ranging from -128 to 127.
  REM
  REM The LCDC register is at FF40h.
  REM Bit 6 - Window tile map address (0: 9800h-9BFFh, 1: 9C00h-9FFFh)
  REM Bit 5 - Window Display (0: off, 1: on)
  REM Bit 4 - Window & background tile data address (0: 8800h-97FFh, 1: 8000h-8FFFh)
  REM Bit 0 - Window & background display (0: off, 1: on)
  REM
  REM The palette for both background and window is located at FF47h (BGP)
  REM
  STATIC AS UBYTE x, y, z, t, UbyteBuff1, UbyteBuff2, Pixrow
  'STATIC AS UBYTE Pixrow
  STATIC AS INTEGER TilesMapAddress, TilesDataAddress, TileNum, TileToDisplay, PixelX, TileTempAddress
  'STATIC AS INTEGER TilesMapAddress, TilesDataAddress
  STATIC TileBufferWin(0 TO 63) AS UBYTE
  IF ((IoRegisters(&hFF40) AND &b00000001) > 0) THEN  ' if background+window bit is enabled at LCDC, then...
    IF ((IoRegisters(&hFF40) AND &b00100000) > 0) THEN  ' if window bit is enabled at LCDC, then...
      IF (IoRegisters(&hFF4B) <= 166) AND (IoRegisters(&hFF4A) <= CurScanline) AND (IoRegisters(&hFF4B) >= 7) AND (CurScanline <= 143) THEN ' if WX and WY are visible, then...

        IF (IoRegisters(&hFF40) AND &b01000000) = 0 THEN
            TilesMapAddress = &h9800
          ELSE
            TilesMapAddress = &h9C00 ' Get tile map address
        END IF
        IF (IoRegisters(&hFF40) AND &b00010000) = 0 THEN
            TilesDataAddress = &h8800
          ELSE
            TilesDataAddress = &h8000 ' Get tile data address
        END IF

        pixrow = (CurScanline - IoRegisters(&hFF4A))
        y = (pixrow SHR 3)  ' don't forget about adding WY! (SHR 3 is the same than / 8)  --> this is the y tile row to display
        pixrow = (pixrow AND &b0111)   ' MOD 8                --> this is the tile's row of pixels to display
        TileNum = (y SHL 5)    ' I'm doing "y SHL 5" instead of "y * 32" for higher speed
        FOR x = 0 TO 31
          IF (TilesDataAddress = &h8000) THEN
              TileToDisplay = VideoRAM(TilesMapAddress + TileNum)
            ELSE
              TileToDisplay = UbyteToByte(VideoRAM(TilesMapAddress + TileNum)) + 128
          END IF
          z = (pixrow SHL 1)   '      Get only the tile's y line related to current scanline
          TileTempAddress = TilesDataAddress + (TileToDisplay SHL 4) + z  ' SHL 4 is the same than *16 (just much faster)
          z = z SHL 2
          'FOR z = 0 TO 15 STEP 2
            UbyteBuff1 = VideoRAM(TileTempAddress)     ' SHL 4 is the same than *16 (just much faster)
            UbyteBuff2 = VideoRAM(TileTempAddress + 1) ' SHL 4 is the same than *16 (just much faster)
            TileBufferWin(z + 7) =  ((UbyteBuff2 AND &b00000001) OR ((UbyteBuff1 AND &b00000001) SHL 1))      ' Note: I am using "SHL 2" here instead of "*4" (faster!)
            TileBufferWin(z + 6) = (((UbyteBuff2 AND &b00000010) SHR 1) OR (((UbyteBuff1 AND &b00000010) SHR 1) SHL 1))
            TileBufferWin(z + 5) = (((UbyteBuff2 AND &b00000100) SHR 2) OR (((UbyteBuff1 AND &b00000100) SHR 2) SHL 1))
            TileBufferWin(z + 4) = (((UbyteBuff2 AND &b00001000) SHR 3) OR (((UbyteBuff1 AND &b00001000) SHR 3) SHL 1))
            TileBufferWin(z + 3) = (((UbyteBuff2 AND &b00010000) SHR 4) OR (((UbyteBuff1 AND &b00010000) SHR 4) SHL 1))
            TileBufferWin(z + 2) = (((UbyteBuff2 AND &b00100000) SHR 5) OR (((UbyteBuff1 AND &b00100000) SHR 5) SHL 1))
            TileBufferWin(z + 1) = (((UbyteBuff2 AND &b01000000) SHR 6) OR (((UbyteBuff1 AND &b01000000) SHR 6) SHL 1))
            TileBufferWin(z) =     (((UbyteBuff2 AND &b10000000) SHR 7) OR (((UbyteBuff1 AND &b10000000) SHR 7) SHL 1))
          'NEXT z
          PixelX = ((x SHL 3) + IoRegisters(&hFF4B) - 7)  ' &hFF4B is the WX register (starts at 7!)   / (x SHL 3) is the same than (x * 8), but faster
          z = ((CurScanline - IoRegisters(&hFF4A)) AND &b0111) SHL 3   ' "AND &b0111" is used instead of "MOD 8" and "SHL 3" instead of "* 8"
          FOR t = 0 TO 7
            IF (PixelX >= 0) AND (PixelX < 160) THEN
              'ScreenBuffer(PixelX, CurScanline) = GbPalette(TileBufferWin(z), 0)
              GbPalette0(TileBufferWin(z), ScreenBuffer(PixelX, CurScanline))
            END IF
            PixelX += 1  ' add 1 instead of relying on t value (faster)
            z += 1
          NEXT t
          TileNum += 1 ' I'm adding 1 instead of relying on x value (faster)
        NEXT x
       END IF
    END IF
  END IF
END SUB


SUB DrawSprites(BYREF CurScanline AS UBYTE)
  STATIC AS UBYTE PatternNum, SpriteFlags, UbyteBuff1, UbyteBuff2, SpritePalette, x, y, z, t
  'STATIC AS INTEGER intX, intY, intZ, intT, SpritePosX, SpritePosY, SpriteMaxY = 7
  'STATIC AS UBYTE PatternNum, SpriteFlags, SpritePalette
  STATIC AS INTEGER SpritePosX, SpritePosY, SpriteMaxY
  STATIC SpriteBuff(0 TO 127) AS UBYTE  ' big enough for 8x8 and 8x16 sprites
  'SpriteFlags = 0
  IF ((IoRegisters(&hFF40) AND &b00000010) > 0) AND (HideSpritesDisplay = 0) THEN ' If bit 1 of LCDC is false -> don't display sprites
    FOR x = 0 TO 39  ' Read 40 blocks of 4 bytes starting from &hFE00
      SpritePosY = SpriteOAM(&hFE00 + (x SHL 2)) - 16     ' sprite's ypos on screen
      SpritePosX = SpriteOAM(&hFE00 + (x SHL 2) + 1) - 8  ' sprite's xpos on screen
      IF (IoRegisters(&hFF40) AND &b00000100) = 0 THEN  ' If Bit 2 of LCDC [FF40] is reset -> the sprite is 8x8
          PatternNum = SpriteOAM(&hFE00 + (x SHL 2) + 2)    ' pattern num
          SpriteMaxY = 7
        ELSE  ' Else the sprite is 8x16 (the LSB is ignored in 8x16 sprite mode)
          PatternNum = (SpriteOAM(&hFE00 + (x SHL 2) + 2) AND &b11111110)
          SpriteMaxY = 15
      END IF
      'REM  /==== DEBUG ====\
      'LOCATE 40+x, 39: PRINT "" & x & ": " & PatternNum & "(" & SpritePosX & "x" & SpritePosY & ")";
      'REM  \==== DEBUG ====/
      IF (SpritePosX >= -8) AND (SpritePosY >= (CurScanline - SpriteMaxY)) AND (SpritePosX < 160) AND (SpritePosY <= CurScanline) THEN
      'IF (SpritePosX >= -8) AND (SpritePosY >= (CurScanline - 20)) AND (SpritePosX < 160) AND (SpritePosY <= CurScanline) THEN  ' This one works fine with Batman - return of the Joker (when Batman jumps there is some flashing whitespace on him otherwise)
          SpriteFlags = SpriteOAM(&hFE00 + (x SHL 2) + 3)        ' Flags
          IF (SpriteFlags AND &b00010000) = 0 THEN
              SpritePalette = 1
            ELSE
              SpritePalette = 2
          END IF
          FOR z = 0 TO ((SpriteMaxY SHL 1) + 1) STEP 2
            UbyteBuff1 = VideoRAM(&h8000 + (PatternNum SHL 4) + z)
            UbyteBuff2 = VideoRAM(&h8000 + (PatternNum SHL 4) + z + 1)
            SpriteBuff((z SHL 2) + 7) =  ((UbyteBuff2 AND &b00000001) OR ((UbyteBuff1 AND &b00000001) SHL 1))
            SpriteBuff((z SHL 2) + 6) = (((UbyteBuff2 AND &b00000010) SHR 1) OR (((UbyteBuff1 AND &b00000010) SHR 1) SHL 1))
            SpriteBuff((z SHL 2) + 5) = (((UbyteBuff2 AND &b00000100) SHR 2) OR (((UbyteBuff1 AND &b00000100) SHR 2) SHL 1))
            SpriteBuff((z SHL 2) + 4) = (((UbyteBuff2 AND &b00001000) SHR 3) OR (((UbyteBuff1 AND &b00001000) SHR 3) SHL 1))
            SpriteBuff((z SHL 2) + 3) = (((UbyteBuff2 AND &b00010000) SHR 4) OR (((UbyteBuff1 AND &b00010000) SHR 4) SHL 1))
            SpriteBuff((z SHL 2) + 2) = (((UbyteBuff2 AND &b00100000) SHR 5) OR (((UbyteBuff1 AND &b00100000) SHR 5) SHL 1))
            SpriteBuff((z SHL 2) + 1) = (((UbyteBuff2 AND &b01000000) SHR 6) OR (((UbyteBuff1 AND &b01000000) SHR 6) SHL 1))
            SpriteBuff((z SHL 2) + 0) = (((UbyteBuff2 AND &b10000000) SHR 7) OR (((UbyteBuff1 AND &b10000000) SHR 7) SHL 1))
          NEXT z
          REM  Here I perform any required transformation on the sprite (vertical and/or horizontal mirroring)
          'IF (SpriteFlags AND &b01000000) > 0 THEN   ' Bit 6 = Y flip (vertical mirror)
          IF (SpriteFlags AND &b00100000) > 0 THEN   ' Bit 5 = X flip (horizontal mirror)
            FOR z = 0 TO SpriteMaxY
              t = SpriteBuff((z SHL 3) + 0)
              SpriteBuff((z SHL 3) + 0) = SpriteBuff((z SHL 3) + 7)
              SpriteBuff((z SHL 3) + 7) = t
              t = SpriteBuff((z SHL 3) + 1)
              SpriteBuff((z SHL 3) + 1) = SpriteBuff((z SHL 3) + 6)
              SpriteBuff((z SHL 3) + 6) = t
              t = SpriteBuff((z SHL 3) + 2)
              SpriteBuff((z SHL 3) + 2) = SpriteBuff((z SHL 3) + 5)
              SpriteBuff((z SHL 3) + 5) = t
              t = SpriteBuff((z SHL 3) + 3)
              SpriteBuff((z SHL 3) + 3) = SpriteBuff((z SHL 3) + 4)
              SpriteBuff((z SHL 3) + 4) = t
            NEXT z
          END IF
          IF (SpriteFlags AND &b01000000) > 0 THEN   ' Bit 6 = Y flip (vertical mirror)
            'IF (SpriteFlags AND &b00100000) > 0 THEN   ' Bit 5 = X flip (horizontal mirror)
            FOR z = 0 TO 7
              IF (SpriteMaxY = 7) THEN  ' Swap 8 pixels high sprites
                  FOR y = 0 TO 24 STEP 8
                    t = SpriteBuff(y + z)
                    SpriteBuff(y + z) = SpriteBuff((56 - y) + z)
                    SpriteBuff((56 - y) + z) = t
                  NEXT y
                ELSE   ' Swap 16 pixels high sprites
                  FOR y = 0 TO 56 STEP 8
                    t = SpriteBuff(y + z)
                    SpriteBuff(y + z) = SpriteBuff((120 - y) + z)
                    SpriteBuff((120 - y) + z) = t
                  NEXT y
              END IF
            NEXT z
          END IF
          REM  Now apply the sprite onscreen...
          GbPalette0(0, UbyteBuff1)  ' Get background color into UbyteBuff1
          'FOR z = 0 TO SpriteMaxY
            z = CurScanline - SpritePosY ' Apply to screen only one line of the sprite (no need to do more)
            FOR t = 0 TO 7
              REM  Update all pixels, but not 0 (which are transparent for a sprite)
              IF ((SpritePosX + t) >= 0) AND ((SpritePosY + z) >= 0) AND ((SpritePosX + t) < 160) AND ((SpritePosY + z) < 144) THEN ' don't try to write outside the screen
                IF (SpriteBuff((z SHL 3) + t) > 0) THEN  ' color 0 is transparent on sprites
                  REM If bit 7 of the sprite's flags is set, then the sprite is "hidden" (prevails only over color 0 of background)
                  IF ((SpriteFlags AND &b10000000) = 0) OR (ScreenBuffer(SpritePosX + t, SpritePosY + z) = UbyteBuff1) THEN ' Sprite's priority over background
                    'ScreenBuffer(SpritePosX + t, SpritePosY + z) = GbPalette(SpriteBuff((z SHL 3) + t), SpritePalette)
                    IF SpritePalette = 1 THEN
                        GbPalette1(SpriteBuff((z SHL 3) + t), ScreenBuffer(SpritePosX + t, SpritePosY + z))
                      ELSE
                        GbPalette2(SpriteBuff((z SHL 3) + t), ScreenBuffer(SpritePosX + t, SpritePosY + z))
                    END IF
                  END IF
                END IF
              END IF
            NEXT t
          'NEXT z
      END IF
    NEXT x
  END IF
END SUB


#MACRO DetectScreenChanges()  ' Parsing all the screen to detect changes that occured since last screen refresh
  'DIM AS UBYTE x, y   ' do not declare x and y - these are already declared by RefreshScreen routines
  LcdModifiedFirstY = 255
  LcdModifiedFirstX = 255
  LcdModifiedLastX = 0
  FOR y = 0 TO 143
    LcdModifiedScanline(y) = 0
    FOR x = 0 TO 159
      IF ScreenBuffer(x, y) <> ScreenOldBuffer(x, y) THEN
        ScreenOldBuffer(x, y) = ScreenBuffer(x, y)
        IF LcdModifiedScanline(y) = 0 THEN
          LcdModifiedScanline(y) = 1
          IF (LcdModifiedFirstY > y) THEN LcdModifiedFirstY = y
          LcdModifiedLastY = y
        END IF
        IF (LcdModifiedFirstX > x) THEN LcdModifiedFirstX = x
        IF (LcdModifiedLastX < x) THEN LcdModifiedLastX = x
      END IF
    NEXT x
  NEXT y
#ENDMACRO


SUB RefreshScreen()
  STATIC AS UINTEGER x, y
  STATIC AS UINTEGER a, b, c, d
  STATIC AS UBYTE ScaleStepX, ScaleStepY
  DetectScreenChanges()

  IF (LcdModifiedFirstY < 255) THEN
    STATIC AS UINTEGER PTR sPtr32  ' Prepare a screen pointer for 32bit color depth
    STATIC AS USHORT PTR sPtr16    ' Prepare a screen pointer for 16bit color depth
    STATIC AS UBYTE PTR sPtr8      ' Prepare a screen pointer for 8bit color depth (also used for 4bit)
    IF (sPtr32 = 0) THEN  ' If this is the first time, then init pointers
      sPtr32 = Screenptr
      sPtr16 = Screenptr
      sPtr8 = Screenptr
    END IF
    SCREENLOCK()
    FOR ScaleStepY = 0 TO (GraphicScaleFactor - 1) STEP FastScale
      a = (GraphicWidth * ScaleStepY) + ScreenBorderRight + (ScreenBorderTop * GraphicWidth)
      FOR y = LcdModifiedFirstY TO LcdModifiedLastY
        IF (LcdModifiedScanline(y) = 1) THEN
          b = (y * GraphicWidth * GraphicScaleFactor) + a
          FOR x = LcdModifiedFirstX TO LcdModifiedLastX
            c = (x * GraphicScaleFactor) + b
            FOR ScaleStepX = 0 TO (GraphicScaleFactor - 1)
              REM Note: It's *much* faster to write into ScreenPTR than rely on PSET
              SELECT CASE AS CONST BitDepth
                CASE 32
                  sPtr32[c + ScaleStepX] = ScreenPalette32(ScreenBuffer(x, y))
                CASE 16
                  sPtr16[c + ScaleStepX] = ScreenPalette16(ScreenBuffer(x, y))
                CASE 8
                  sPtr8[c + ScaleStepX] = ScreenBuffer(x, y)
                CASE 4
                  sPtr8[c + ScaleStepX] = ScreenBuffer(x, y)
              END SELECT
            NEXT ScaleStepX
          NEXT x
        END IF
      NEXT y
    NEXT ScaleStepY
    'SCREENSYNC()   -->  serious framerate loss when activated! (but no tearing anymore...)
    SCREENUNLOCK()
  END IF
END SUB


SUB ClearScreen()
  DIM AS UINTEGER x, y, a, b, c
  DIM AS UBYTE ScaleStepX, ScaleStepY
  DIM AS UINTEGER PTR sPtr32 = Screenptr ' Prepare a screen pointer for 32bit color depth
  DIM AS USHORT PTR sPtr16 = Screenptr   ' Prepare a screen pointer for 16bit color depth
  DIM AS UBYTE PTR sPtr8 = Screenptr     ' Prepare a screen pointer for 8bit color depth (also used for 4bit)
  SCREENLOCK()
  FOR ScaleStepY = 0 TO (GraphicScaleFactor - 1)
    a = (GraphicWidth * ScaleStepY) + ScreenBorderRight + (ScreenBorderTop * GraphicWidth)
    FOR y = 0 TO 143
      b = (y * GraphicWidth * GraphicScaleFactor) + a
      FOR x = 0 TO 159
        c = (x * GraphicScaleFactor) + b
        FOR ScaleStepX = 0 TO (GraphicScaleFactor - 1)
          REM Note: It's *much* faster to write into ScreenPTR than rely on PSET
          SELECT CASE AS CONST BitDepth
            CASE 32
              sPtr32[c + ScaleStepX] = ScreenPalette32(15)
            CASE 16
              sPtr16[c + ScaleStepX] = ScreenPalette16(15)
            CASE 8
              sPtr8[c + ScaleStepX] = 15
            CASE 4
              sPtr8[c + ScaleStepX] = 15
          END SELECT
        NEXT ScaleStepX
        ScreenBuffer(x, y) = 15     ' don't forget to sync internal screen buffers,
        ScreenOldBuffer(x, y) = 15  ' otherwise LCD change detection won't work correctly!
      NEXT x
    NEXT y
  NEXT ScaleStepY
  SCREENUNLOCK()
END SUB


SUB TurnLcdOff()
  DIM AS UBYTE x, y
  FOR y = 0 TO 143
    FOR x = 0 TO 159
      ScreenBuffer(x,y) = 0
    NEXT x
  NEXT y
END SUB


SUB InitScreenOldBuffer() ' Call this one time, at the emulator
  DIM AS UBYTE x, y       ' start, to force the refresh of the
  FOR y = 0 TO 143        ' whole first LCD frame.
    FOR x = 0 TO 159
      ScreenOldBuffer(x,y) = 255 ' fill with an unlikely value
    NEXT x
  NEXT y
END SUB


#DEFINE GetLcdMode() (IoRegisters(&hFF41) AND &b00000011)

#MACRO SetLcdMode(x)
  IoRegisters(&hFF41) = (IoRegisters(&hFF41) AND &b11111100)
  IoRegisters(&hFF41) = (IoRegisters(&hFF41) OR x)   ' Here I set the LCD mode
#ENDMACRO


#MACRO VideoSysUpdate(cycles)
'SUB VideoSysUpdate(cycles AS UBYTE)
  VideoClkCounterVBlank += cycles
  IF VideoClkCounterVBlank >= 70224 THEN
    VideoClkCounterVBlank -= 70224
    VideoClkCounterMode = VideoClkCounterVBlank ' Sync VideoClkCounterMode with VideoClkCounterVBlank
  END IF

  REM  MemoryInternalHiRAM(&hFF44) is the LY register (current vertical scanline)

  IF (IoRegisters(&hFF40) AND &b10000000) > 0 THEN   ' If LCD is ON...
      CurLY = (VideoClkCounterVBlank \ 456)        ' LY should be between 0..153

      IF (CurLY <> LastLY) THEN                    ' If LY changed, then...
        LastLY = CurLY
        IoRegisters(&hFF44) = CurLY                ' Save new LY
      END IF

      IF IoRegisters(&hFF44) = IoRegisters(&hFF45) THEN         ' Compare LY with LYC, and if equal,
          IoRegisters(&hFF41) = (IoRegisters(&hFF41) OR &b100)  ' set the coincidence (2) bit of the
          IF (IoRegisters(&hFF41) AND &b01000000) > 0 THEN      ' STAT register (FF41) and (if enabled
            IoRegisters(&hFF0F) = (IoRegisters(&hFF0F) OR &b10) ' via bit 6 of STAT register) trigger
          END IF                                                ' the Coincidence (LCDC) interrupt.
        ELSE
          IoRegisters(&hFF41) = (IoRegisters(&hFF41) AND &b11111011) ' reset the coincidence flag
      END IF
    ELSE
      IoRegisters(&hFF44) = 0 ' LY = 0 when LCD is off
  END IF


  IF (VideoClkCounterVBlank >= 65664) THEN ' We are in a VBLANK period!
      REM Here I trigger the vblank interrupt and set mode 1
      IF GetLcdMode() <> &b01 THEN   ' Check if not already in mode 1
        SetLcdMode(1)   ' Here I set LCD mode 1
        IoRegisters(&hFF0F) = (IoRegisters(&hFF0F) OR &b00000001)   ' Ask for the VBLANK interrupt
        IF (SkipFramesCounter = SkipFrames) THEN
          REM Check if a screenshot was requested...
          IF ((AskForScreenshot = 1) AND ((TIMER - LastScreenshotTime) > 1)) OR ((AutoScreenshot = 1) AND ((TIMER - LastScreenShotTime) >= 10)) THEN
              DIM AS DOUBLE TempDate
              TempDate = NOW
              SavePcxFile(LEFT(RomInfos.Filename, INSTRREV(RomInfos.Filename, ".") - 1) & "_" & YEAR(NOW) & "-" & RIGHT("00" & MONTH(NOW), 2) & "-" & RIGHT("00" & DAY(NOW), 2) & "_" & RIGHT("00" & HOUR(NOW), 2) & RIGHT("00" & MINUTE(NOW), 2) & RIGHT("00" & SECOND(NOW), 2) & ".pcx")
              AskForScreenshot = 0
              LastScreenShotTime = TIMER
              UserMessage = "SCREENSHOT DONE!"
          END IF
          IF (IoRegisters(&hFF40) AND &b10000000) = 0 THEN ' The LCD is off, make it all black
            TurnLcdOff()
            'UserMessage = "LCD IS OFF"
          END IF
          AdjustTiming()               ' Slow down CPU emulation before drawing graphic frame (if running too fast)
          IF LEN(UserMessage) > 0 THEN
              PrintMsg(UserMessage, 1) ' do not use refreshscreen() here, (it is called by PrintMsg anyway)
            ELSE
              RefreshScreen()
          END IF
          SkipFramesCounter = 0
          CountFPS += 1
        END IF
        SkipFramesCounter += 1
      END IF
    ELSE ' Outside the VBlank, perform the MODE 0-2-3 loop
      VideoClkCounterMode += cycles
      IF VideoClkCounterMode >= 456 THEN
        VideoClkCounterMode -= 456
      END IF
      SELECT CASE VideoClkCounterMode
        CASE IS <= 80   ' mode 2
        'CASE 0
          IF GetLcdMode() <> 2 THEN   '  Check if not already in mode 2
            SetLcdMode(2)   ' Here I set LCD mode 2
            IF (IoRegisters(&hFF41) AND &b00100000) > 0 THEN   ' If OAM int is enabled, request LCDC int
              IoRegisters(&hFF0F) = (IoRegisters(&hFF0F) OR &b00000010)
            END IF
          END IF
        CASE IS <= 172  ' mode 3
        'CASE 1, 2
          IF GetLcdMode() <> 3 THEN  ' Check if not already in mode 3
            SetLcdMode(3)   ' Here I set LCD mode 3 (no int here)

            IF (IoRegisters(&hFF40) AND &b10000000) > 0 THEN   ' If LCD is ON...
              IF LastLYdraw <> CurLY THEN                      ' And curline hasn't been drawn yet...
                LastLYdraw = CurLY
                IF (SkipFramesCounter = SkipFrames) THEN
                  'IF (CurLY < 144) THEN
                    DrawBackground(CurLY)                      ' Generate current scanline
                    'DrawLine(CurLY)
                    DrawWindow(CurLY)
                    DrawSprites(CurLY)
                  'END IF
                END IF
              END IF
            END IF

          END IF
        CASE ELSE       ' mode 0
        'CASE 3, 4, 5
          IF GetLcdMode() <> 0 THEN  ' Check if not already in mode 0
            ' Trigger the hblank interrupt if enabled via bit 3 of the stat register (FF41) (via LCDC int?)
            SetLcdMode(0)   ' Here I set LCD mode 0
            IF (IoRegisters(&hFF41) AND &b00001000) > 0 THEN   ' If hblank int is enabled, request LCDC int
              IoRegisters(&hFF0F) = (IoRegisters(&hFF0F) OR &b00000010)
            END IF
          END IF
      END SELECT



      'IF (VideoClkCounterMode <= 201) THEN    ' Just mode 0
      '  REM Mode 0, do nothing  --  Don't request INT! Already done when passed from mode 3 to mode 0!
      '  'IoRegisters(&hFF0F) = (IoRegisters(&hFF0F) OR &b00000010)   ' Ask for the LCDC interrupt via IF (FF0F)
      'ELSEIF (VideoClkCounterMode <= 278) THEN                          ' MODE 2 (77..83 clks / 19us) generate OAM interrupt if enabled via bit 5 of STAT (via LCDC int?)
      '  IF GetLcdMode() <> 2 THEN   '  Check if not already in mode 2
      '    SetLcdMode(2)   ' Here I set LCD mode 2
      '    IF (IoRegisters(&hFF41) AND &b00100000) > 0 THEN   ' If OAM int is enabled, request LCDC int
      '      IoRegisters(&hFF0F) = (IoRegisters(&hFF0F) OR &b00000010)
      '    END IF
      '  END IF
      ''ELSEIF (VideoClkCounterMode <= 447) THEN                       ' MODE 3 (169..175 clks / 41us)
      'ELSEIF (VideoClkCounterMode <= 456) THEN                       ' MODE 3 (169..175 clks / 41us)
      '  SetLcdMode(3)   ' Here I set LCD mode 3 (no int here)
      'ELSE  ' Mode 0 (201..207 clks / 48.6us) and wrap VideoClkCounterMode
      '  ' Trigger the hblank interrupt if enabled via bit 3 of the stat register (FF41) (via LCDC int?)
      '  SetLcdMode(0)   ' Here I set LCD mode 0
      '  VideoClkCounterMode -= 456
      '  IF (IoRegisters(&hFF41) AND &b00001000) > 0 THEN   ' If hblank int is enabled, request LCDC int
      '    IoRegisters(&hFF0F) = (IoRegisters(&hFF0F) OR &b00000010)
      '  END IF
      'END IF
  END IF
#ENDMACRO
'END SUB
